/*
Copyright (c) 2011, Honeywell International Inc.  Developed with the
sponsorship of the Defense Advanced Research Projects Agency DARPA .

Permission is hereby granted, free of charge, to any person obtaining
a copy of this data, including any software or models in source or
binary form, as well as any drawings, specifications, and
documentation collectively "the Data", to deal in the Data without
restriction, including without limitation the rights to use, copy,
modify, merge, publish, distribute, sublicense, and/or sell copies of
the Data, and to permit persons to whom the Data is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Data .

THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS,
CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR
OTHER DEALINGS IN THE DATA.
*/

//###############################################################################################################################################
//
//	PromiseVisitor.cpp
//
//###############################################################################################################################################

#include "stdafx.h"
#include "PromiseVisitor.h"

namespace BON
{
const char* PromiseVisitor::ATTRIBUTE = "Attribute";
const char* PromiseVisitor::COMPOSITE = "Composite";
const char* PromiseVisitor::DATACONNECTION = "DataConnection";
const char* PromiseVisitor::DTD = "<!DOCTYPE Network SYSTEM \"network.dtd\">";
const char* PromiseVisitor::ENDNETWORK = "</Network>";
const char* PromiseVisitor::ENDSYSTEM = "EndSystem";
const char* PromiseVisitor::FAULT = "Fault";
const char* PromiseVisitor::HOST = "Host";
const char* PromiseVisitor::ID = "Id";
const char* PromiseVisitor::NETWORK = "<Network";
const char* PromiseVisitor::PATH = "Path";
const char* PromiseVisitor::SWITCH = "Switch";
const char* PromiseVisitor::TOVIRTUALLINK = "ToVirtualLink";
const char* PromiseVisitor::TRAFFICTYPE = "TrafficType";
const char* PromiseVisitor::VALUE = "Value";
const char* PromiseVisitor::VIRTUALCONNECTION = "VirtualConnection";
const char* PromiseVisitor::VIRTUALLINK = "VirtualLink";
const char* PromiseVisitor::XMLHEADER = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
PromiseVisitor* PromiseVisitor::_instance = NULL;

// Operator overloading
std::string& operator<< (std::string& out, const std::string& parameter)
{
	out += parameter;
	return out;
}
std::string& operator<< (std::string& out, const char* parameter)
{
	out += parameter;
	return out;
}
std::string& operator<< (std::string& out, int parameter)
{
	char number[char_size];
	sprintf_s(number, char_size, "%d", parameter);
	out += number;
	return out;
}

// Promise member classes
PromiseVisitor::PromiseVisitor(Dialog* dlg, const std::string& fileName)
	: _dlg(dlg)
{
	init(fileName);
}
PromiseVisitor::~PromiseVisitor()
{
	_outStream->flush();
	_outStream->close();
	delete _outStream;
}
Model PromiseVisitor::getConnDst(const Connection& connection)
{
	Atom dstAtom = connection->getDst();
	assert(NULL != dstAtom);
	Model dstModel = dstAtom->getParentModel();
	assert(NULL != dstModel);
	// Check whether destination is composite type
	MON::FCO fco = dstModel->getFCOMeta();
	if (0 == fco.name().compare(COMPOSITE))
	{
		// It is Composite, need to continue resolving...
		// Find destination atom outgoing connection
		//Connection outConn;
		std::set<Connection> connSet = dstAtom->getOutConnLinks(DATACONNECTION);
		// Should be only 1 outgoing connection...
		assert(connSet.size() == 1);
		std::set<Connection>::const_iterator iter = connSet.begin();
		return getConnDst(*iter);
	} else {
		// Found destination model
		return dstModel;
	}
}
Model PromiseVisitor::getConnSrc(const Connection& connection)
{
	Atom srcAtom = connection->getSrc();
	assert(NULL != srcAtom);
	Model srcModel = srcAtom->getParentModel();
	assert(NULL != srcModel);
	// Check whether destination is composite type
	MON::FCO fco = srcModel->getFCOMeta();
	if (0 == fco.name().compare(COMPOSITE))
	{
		// It is Composite, need to continue resolving...
		// Find destination atom outgoing connection
		//Connection outConn;
		std::set<Connection> connSet = srcAtom->getInConnLinks(DATACONNECTION);
		// Should be only 1 incoming connection...
		assert(connSet.size() == 1);
		std::set<Connection>::const_iterator iter = connSet.begin();
		return getConnSrc(*iter);
	} else {
		// Found destination model
		return srcModel;
	}
}
void PromiseVisitor::getAttributes(const Model& model, std::map<std::string, std::string>& attributeMap,
	std::map<std::string, std::string>& faultMap)
{
	std::set<Atom> childAtoms = model->getChildAtoms();
	for (std::set<Atom>::const_iterator iter = childAtoms.begin(); iter != childAtoms.end(); ++iter)
	{
		// Check whether atom is Attribute or Fault
		Atom atom = *iter;
		MON::FCO meta = atom->getFCOMeta();
		if (0 == meta.name().compare(FAULT))
		{
			faultMap.insert(std::pair<std::string, std::string>(atom->getName(),
				atom->getAttribute(VALUE)->getStringValue()));
		}
		else if (0 == meta.name().compare(ATTRIBUTE))
		{
			attributeMap.insert(std::pair<std::string, std::string>(atom->getName(),
				atom->getAttribute(VALUE)->getStringValue()));
		}
	}
}
Model PromiseVisitor::getHostDst(const Model& model)
{
	std::set<Atom> ports = model->getChildAtoms();
	for (std::set<Atom>::const_iterator iter = ports.begin(); iter != ports.end(); ++iter)
	{
		// Found a port. Let's get destination.
		std::set<Connection> outConns = (*iter)->getOutConnLinks();
		for (std::set<Connection>::const_iterator iter2 = outConns.begin(); iter2 != outConns.end(); ++iter2)
		{
			Atom dstPort = (*iter2)->getDst();
			Model dstModel = dstPort->getParent();
			MON::FCO meta = dstModel->getFCOMeta();
			if (0 == meta.name().compare(HOST))
			{
				// Found host, return it
				return dstModel;
			}
		}
	}
	// If we're here, the model is incorrect.
	return model;
}
Model PromiseVisitor::getHostSrc(const Model& model)
{
	std::set<Atom> ports = model->getChildAtoms();
	for (std::set<Atom>::const_iterator iter = ports.begin(); iter != ports.end(); ++iter)
	{
		// Found a port. Let's get source.
		std::set<Connection> inConns = (*iter)->getInConnLinks();
		for (std::set<Connection>::const_iterator iter2 = inConns.begin(); iter2 != inConns.end(); ++iter2)
		{
			Atom srcPort = (*iter2)->getSrc();
			Model srcModel = srcPort->getParent();
			MON::FCO meta = srcModel->getFCOMeta();
			if (0 == meta.name().compare(HOST))
			{
				// Found host, return it
				return srcModel;
			}
		}
	}
	// If we're here, the model is incorrect.
	return model;
}
void PromiseVisitor::getPath(std::string str, std::vector<int>& pathVector)
{
	int x = str.find(",");
	while (x < str.size())
	{
	  // Find number
	  std::string subStr = str.substr(0, x);
	  // Truncate original string
	  str = str.substr(x + 1);
	  int number = atoi(subStr.c_str());
	  pathVector.push_back(number);
	  // Next element
      x = str.find(",");
	}
    int number = atoi(str.c_str());
	pathVector.push_back(number);
}
void PromiseVisitor::getPathLine(std::string str, std::vector<std::string>& pathLine)
{
	int x = str.find("\n");
	while (x < str.size())
	{
	  // Find number
	  std::string subStr = str.substr(0, x);
	  // Truncate original string
	  str = str.substr(x + 1);
	  pathLine.push_back(subStr);
	  // Next element
      x = str.find("\n");
	}
}
void PromiseVisitor::getVirtualLinkDst(const FCO& fco, std::map<int, Model>& models)
{
	std::set<Connection> outConns = fco->getOutConnLinks();
	// Incoming connections
	for (std::set<Connection>::const_iterator iter = outConns.begin(); iter != outConns.end(); ++iter)
	{
		Atom dstPort = (*iter)->getDst();
		Model model = dstPort->getParent();
		MON::FCO meta = model->getFCOMeta();
		if (0 == meta.name().compare(COMPOSITE))
		{
			getVirtualLinkDst(dstPort, models);
		}
		else
		{
			// Need to resolve endsystem
			Model host = getHostDst(model);
			models.insert(std::pair<int, Model>(host->getAttribute(ID)->getIntegerValue(), host));
		}
	}
}
Model PromiseVisitor::getVirtualLinkSrc(const FCO& fco)
{
	std::set<Connection> inConns = fco->getInConnLinks();
	Model model = NULL;
	Atom srcPort;
	// Incoming connections
	for (std::set<Connection>::const_iterator iter = inConns.begin(); iter != inConns.end(); ++iter)
	{
		srcPort = (*iter)->getSrc();
		if (NULL == model)
		{
			// EndSystem
			model = srcPort->getParent();
		}
		else
		{
			// In case of multiple in connections, they should all come from the same model
			assert(model == srcPort->getParent());
		}
	}
	MON::FCO meta = model->getFCOMeta();
	if (0 == meta.name().compare(COMPOSITE))
	{
		return getVirtualLinkSrc(srcPort);
	}
	model = getHostSrc(model);
	return model;
}
void PromiseVisitor::init(const std::string& fileName)
{
	_dlg->m_strValue = "";
	// Open output file.
	_outStream = new std::ofstream(fileName.c_str(), std::ios::out);
}
PromiseVisitor* PromiseVisitor::instance(Dialog* dlg, const std::string& fileName)
{
	assert(NULL != dlg);
	if (NULL == _instance)
	{
		_instance = new PromiseVisitor(dlg, fileName);
		assert(NULL != _instance);
	}
	return _instance;
}
void PromiseVisitor::visit(const Folder& folder)
{
	// Write XML header information.
	writeLine(XMLHEADER);
	writeLine(DTD);
	std::set<Model> childModels = folder->getChildModels();
	// Visit child models recursively.
	for (std::set<Model>::iterator iter = childModels.begin(); iter != childModels.end(); ++iter)
	{
		assert (NULL != (*iter));
		visit(*iter);
	}
}
void PromiseVisitor::visit(const Model& model)
{
	// Get attributes and fault list
	std::map<std::string, std::string> attributeMap;
	std::map<std::string, std::string> faultMap;
	getAttributes(model, attributeMap, faultMap);
	std::string params;
	params << NETWORK;
	for (std::map<std::string, std::string>::const_iterator iter = attributeMap.begin(); iter != attributeMap.end(); ++iter)
	{
		params << " " << iter->first << "=\"" << iter->second << "\"";
	}
	params << ">";
	writeLine(params);
	// Process network actors
	std::map<int, Model> actors;
	std::set<Connection> connections;
	std::map<int, Atom> virtualLinks;
	visitActors(model, actors, connections, virtualLinks);
	std::string str;
	// Wiriting component list
	str << "  <ComponentList>";
	writeLine(str);
	int x = 120;
	int y = 80;
	// Enumerating components
	for (std::map<int, Model>::const_iterator iter = actors.begin(); iter != actors.end(); ++iter)
	{
		// Write actors to XML
		MON::FCO fcoMeta = iter->second->getFCOMeta();
		str = "";
		str << "    <Component id=\"" << iter->first << "\" type=\"" << fcoMeta.name() << "\" integrity=\""
			<< iter->second->getAttribute("Integrity")->getStringValue() << "\" name=\"" << iter->second->getName()
			<< "\" x=\"" << x << "\" y = \"" << y << "\"/>";
		x = (500 < x) ? 120 : x + 120;
		y = (500 < x) ? y + 80 : y;
		writeLine(str);
	}
	str = "";
	str << "  </ComponentList>";
	writeLine(str);
	str = "";
	// Writing connection list
	str << "  <ConnectionList>";
	writeLine(str);
	std::set<std::pair<int, int> > connSet;
	// Filtering connections
	for (std::set<Connection>::const_iterator iter = connections.begin(); iter != connections.end(); ++iter)
	{
		Model dstModel = getConnDst(*iter);
		Model srcModel = getConnSrc(*iter);
		// This should remove redundant connections
		int id1 = srcModel->getAttribute(ID)->getIntegerValue();
		int id2 = dstModel->getAttribute(ID)->getIntegerValue();
		if (id1 < id2) {
			std::pair<int, int> connIdPair(id1, id2);
			if (connSet.find(connIdPair) == connSet.end())
			{
				connSet.insert(connIdPair);
			}
		} else {
			std::pair<int, int> connIdPair(id2, id1);
			if (connSet.find(connIdPair) == connSet.end())
			{
				connSet.insert(connIdPair);
			}
		}
	}
	int i=0;
	// Enumerating connections
	for(std::set<std::pair<int,int> >::const_iterator iter = connSet.begin(); iter != connSet.end(); ++iter)
	{
		// Write connections to XML
		str = "";
		std::string up = "";
		up << "\" upstream=\"" << ++i;
		std::string down = "";
		down << "\" downstream=\"" << ++i;
		str << "    <Connection source=\"" << iter->first << "\" sink=\"" << iter->second << up << down << "\"/>";
		writeLine(str);
	}
	str = "";
	str << "  </ConnectionList>";
	writeLine(str);
	// Writing virtual link list
	str = "";
	str << "  <VirtualLinkList>";
	writeLine(str);
	// Enumerating virtual links
	for(std::map<int, Atom>::const_iterator iter = virtualLinks.begin(); iter != virtualLinks.end(); ++iter)
	{
		// Write virtual links to XML
		str = "";
		str << "    <VirtualLink id=\"" << iter->first << "\" channels=\"" << iter->second->getInConnLinks().size()
			<< "\" frameLength=\"" << iter->second->getAttribute("FrameLength")->getStringValue() << "\">";
		writeLine(str);
		// Get virtual connection sources
		str = "";
		Model model = getVirtualLinkSrc(iter->second);
		str << "      <Source id=\"" << model->getAttribute(ID)->getStringValue() << "\"/>";
		writeLine(str);
		// Get virtual connection destinations
		str = "";
		str << "      <DestinationList>";
		writeLine(str);
		std::map<int, Model> destinations;
		getVirtualLinkDst(iter->second, destinations);
		// Iterate through destinations
		for (std::map<int, Model>::const_iterator dstIter = destinations.begin(); dstIter != destinations.end(); ++dstIter)
		{
			str = "";
			str << "        <Destination id=\"" << dstIter->first << "\"/>";
			writeLine(str);
		}
		str = "";
		str << "      </DestinationList>";
		writeLine(str);
		// Traffic type
		str = "";
		str << "      <Traffic type=\"" << iter->second->getAttribute(TRAFFICTYPE)->getStringValue() << "\"/>";
		int trafficType = 0;
		if (0 == iter->second->getAttribute(TRAFFICTYPE)->getStringValue().compare("TT"))
		{
			trafficType = 1;
		}
		writeLine(str);
		// Get paths
		str = "";
		str << "      <PathList>";
		writeLine(str);
		std::string path = iter->second->getAttribute(PATH)->getStringValue();
		std::vector<std::string> pathLine;
		getPathLine(path, pathLine);
		// Iterate through paths
		int channel = 0;
		int skipDirection = 0;
		std::map<int, Model>::const_iterator destination = destinations.begin();
		for (std::vector<std::string>::const_iterator iter = pathLine.begin(); iter != pathLine.end(); ++ iter)
		{
			// Iterate through elements within paths
			std::vector<int> pathVector;
			getPath(*iter, pathVector);
			str = "";
			str << "        <Path channel=\"" << channel << "\" destination=\"" << destination->first << "\" route=\"";
			if (0 == skipDirection)
			{
				skipDirection = 1;
			}
			else
			{
				skipDirection = 0;
				if (destination != destinations.end())
				{
				  ++destination;
				}
			}
//			if (trafficType)
//			{
			channel = (0 == channel) ? 1 : 0;
//			}
			int first = 1;
			for (std::vector<int>::const_iterator iter2 = pathVector.begin(); iter2 != pathVector.end(); ++iter2)
			{
				if (first) {
				  str << (*iter2);
				  first = 0;
				}
				else
				{
				  str << "," << (*iter2);
				}
			}
			str << "\"/>";
			writeLine(str);
		}
		str = "";
		str << "      </PathList>";
		writeLine(str);
		// Closing tag
		str = "";
		str << "    </VirtualLink>";
		writeLine(str);
	}
	str = "";
	str << "  </VirtualLinkList>";
	writeLine(str);
	// Write fault list
	str = "";
	str << "  <FaultList>";
	writeLine(str);
	for (std::map<std::string, std::string>::const_iterator iter = faultMap.begin(); iter != faultMap.end(); ++iter)
	{
		str ="";
		str << "    <Fault type=\"" << iter->first << "\" rate=\"" << iter->second << "\"/>";
		writeLine(str);
	}
	str = "";
	str << "  </FaultList>";
	writeLine(str);
	writeLine(ENDNETWORK);
}
void PromiseVisitor::visitActors(const Model& model, std::map<int, Model>& actors,
	std::set<Connection>& connections, std::map<int, Atom>& virtualLinks)
{
	std::set<Model> childModels = model->getChildModels();
	for (std::set<Model>::const_iterator iter = childModels.begin(); iter != childModels.end(); ++iter)
	{
		// Check whether model is switch, endsystem, host, or composite
		MON::FCO fco = ((*iter)->getFCOMeta());
		if (0 == fco.name().compare(COMPOSITE))
		{
			visitActors(*iter, actors, connections, virtualLinks);
		}
		// Add network actors to set
		if((0 == fco.name().compare(SWITCH)) || (0 == fco.name().compare(ENDSYSTEM)) || (0 == fco.name().compare(HOST)))
		{
			// Add object to list
			actors.insert(std::pair<int, Model>((*iter)->getAttribute(ID)->getIntegerValue(), *iter));
		}
	}
	// Process connections
	std::set<Connection> childConnections = model->getChildConnections();
	for (std::set<Connection>::const_iterator iter = childConnections.begin(); iter != childConnections.end(); ++iter)
	{
		// Check whether connection is data connection
		MON::FCO fco = ((*iter)->getFCOMeta());
		// Add connections to set
		if(0 == fco.name().compare(DATACONNECTION))
		{
			// Add object to list
			connections.insert(*iter);
		}
	}
	// Process virtual links
	std::set<Atom> childAtoms = model->getChildAtoms();
	for (std::set<Atom>::const_iterator iter = childAtoms.begin(); iter != childAtoms.end(); ++iter)
	{
		// Check whether atom is a virtual link
		MON::FCO fco = ((*iter)->getFCOMeta());
		// Add virtual links to set
		if(0 == fco.name().compare(VIRTUALLINK))
		{
			// Add object to list
			virtualLinks.insert(std::pair<int, Atom>((*iter)->getAttribute(ID)->getIntegerValue(), *iter));
		}
	}
}
void PromiseVisitor::writeLine(const std::string& str)
{
	(*_outStream) << str + "\n";
	_outStream->flush();
	// Enable for debugging
	// _dlg->m_strValue += CString(str.c_str()) + "\r\n";
}
}; // namespace BON
