/*
Copyright (c) 2011, Honeywell International Inc.  Developed with the
sponsorship of the Defense Advanced Research Projects Agency DARPA .

Permission is hereby granted, free of charge, to any person obtaining
a copy of this data, including any software or models in source or
binary form, as well as any drawings, specifications, and
documentation collectively "the Data", to deal in the Data without
restriction, including without limitation the rights to use, copy,
modify, merge, publish, distribute, sublicense, and/or sell copies of
the Data, and to permit persons to whom the Data is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Data .

THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS, SPONSORS, DEVELOPERS,
CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE DATA OR THE USE OR
OTHER DEALINGS IN THE DATA.
*/

//###############################################################################################################################################
//
//	PromiseVisitor.h
//
//###############################################################################################################################################

#ifndef PromiseVisitor_h
#define PromiseVisitor_h

#include "BON.h"
#include "BONImpl.h"
#include "Dialog.h"
#include "Extensions.h"
#include "MON.h"
#include "MONImpl.h"
#include <ComponentConfig.h>
#include <assert.h>
#include <fstream>
#include <iterator>
#include <iostream>

#define char_size 32

namespace BON
{
class PromiseVisitor
{
	public:
		// Defining visitor as singleton class.
		static PromiseVisitor* instance(Dialog* dlg, const std::string& fileName);
		// Destructor.
		~PromiseVisitor();

		// Root folder processing entry point.
		void visit(const Folder& folder);
		void visit(const Model& model);

	private :
		// Private constructor ensures creation through instance() only.
		PromiseVisitor(Dialog* dlg, const std::string& fileName);
		// Resolves connection destination, walking through the model and hierarchy.
		Model getConnDst(const Connection& connection);
		// Resolves connection source, walking through the model and hierarchy.
		Model getConnSrc(const Connection& connection);
		// Fills attribute and fault map.
		void getAttributes(const Model& model, std::map<std::string, std::string>& attributeMap,
			std::map<std::string, std::string>& faultMap);
		// Finds data connection destination that is a Host.
		Model getHostDst(const Model& model);
		// Finds data connection source that is a Host.
		Model getHostSrc(const Model& model);
		// Parses parameter string, and converts it into sorted set of integers.
		void getPath(std::string str, std::vector<int>& pathVector);
		// Breaks path string into lines of strings.
		void getPathLine(std::string str, std::vector<std::string>& pathLine);
		// Resolves virtual link destination
		void getVirtualLinkDst(const FCO& fco, std::map<int, Model>& models);
		// Resolves virtual link source
		Model getVirtualLinkSrc(const FCO& fco);
		// Initialize visitor, open file stream.
		void init(const std::string& fileName);
		// Fill list of actors by visiting the model.
		void visitActors(const Model& model, std::map<int, Model>& actors, std::set<Connection>& connections,
			std::map<int, Atom>& virtualLinks);
		// Print information.
		void writeLine(const std::string& str);

		// Dialog used for displaying information.
		Dialog* _dlg;
		// Output file.
		std::ofstream* _outStream;
		// Visitor instance.
		static PromiseVisitor* _instance;
		// XML header.
		static const char* ATTRIBUTE;
		static const char* COMPOSITE;
		static const char* DATACONNECTION;
		static const char* DTD;
		static const char* ENDNETWORK;
		static const char* ENDSYSTEM;
		static const char* FAULT;
		static const char* HOST;
		static const char* ID;
		static const char* NETWORK;
		static const char* PATH;
		static const char* SWITCH;
		static const char* TOVIRTUALLINK;
		static const char* TRAFFICTYPE;
		static const char* VALUE;
		static const char* VIRTUALCONNECTION;
		static const char* VIRTUALLINK;
		static const char* XMLHEADER;
};

}; // namespace BON

#endif // PromiseVisitor_h